/*
 * Copyright 2005, 2006 University of Leiden.
 *
 * This file is part of MIA+EWS.
 *
 * MIA+EWS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA+EWS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <math.h>
#include "oirSupport.h"
#include "fitsio.h"
#include "imaging_data.h"
#include "imaging_fdata.h"
#include "minrtsFrameData.h"
#include "minrtsFftUtil.h"

#define yBottom 5
#define yTop    34
#define NBIN 5
#define DEBUG 1
#define SIGMAY 2.0
#define RECAL 1
/**********************************************************
 * Usage:
 *    oirSciPhotCal "data" coordFile crosscoeff outFile outMaskFile sigmaY
 *
 *    inFiles is a blank-separated list of input files
 *    coordFile is a fits file containing at least a detector def file
 *      defining the wavelength/distortion coefficients
 *
 *    crosscoeff is a crosscoefficient (kappa coefficient) file name
 *
 *    outFile is the name of the output file
 *
 *    outMaskFile contains the name of a file to contain the
 *      photometric mask (to be used in fringe estimation program)
 *
 *    sigmaY  sigma of masks in Y-direction (pixels) (optional)
 ************************************************************* */
minrtsFrameData   *midiCreateMask(minrtsFrameData *Target, float sigmaY) {

   minrtsFrameData *mask;
   int iy, ix, nx, ny ;
   int i,j, ierr;;
   int iregion;

   float *sumX, *sumY, *yy;
   float y, yC[4], maxx;

   nx = Target->nx[0];
   ny = Target->ny[0];

   sumX = (float *)malloc(ny*sizeof(float));
   sumY = (float *)malloc(nx*sizeof(float));
   yy   = (float *)malloc(nx*sizeof(float));
/*
      loop over PA and PB
      first sum over (most of) Y to find limits in X
      them sum over X to find peak row
*/
   for (i=0;i<ny;i++) yy[i] = (float)i;
   for (i=0; i<4; i=i+3) {
       for (ix=0;ix<nx;ix++) {
          sumY[ix] = 0.;
          for (iy=5; iy<ny-5; iy++) sumY[ix] = sumY[ix] + 
             Target->data[i][iy*nx+ix];
       }
       maxx = -1.e9;
       for (ix=0;ix<nx;ix++) if (sumY[ix] > maxx) maxx = sumY[ix];
       maxx = 0.1 * maxx;
       for (iy=0; iy<ny; iy++) {
          sumX[iy] = 0.;
          for (ix=0;ix<nx;ix++) if (sumY[ix] > maxx) sumX[iy] = sumX[iy] + 
             Target->data[i][iy*nx+ix];
       }
       y = minrtsQPk(ny, yy, sumX, &maxx, 0);
       yC[i] = y;
   }
/*
      average peak of channels 1 and 4
*/
   yC[0] = 0.5*(yC[0]+yC[3]);
      
/*
      create empty object
*/
   mask = minrtsFrameDataClone(Target);
/*
      create gaussian masks that follow curve in coordinate system
*/
   sigmaY = 0.5/(sigmaY*sigmaY);
   for (iregion=0; iregion<4; iregion++) {
      for (ix=0;ix<nx;ix++) {
         for (iy=0;iy<ny;iy++) {
            maxx = sigmaY * (iy-yC[0])*(iy-yC[0]);
            if (maxx < 5.) maxx = exp(-maxx); else maxx = 0.;
            mask->data[iregion][ix+nx*iy] = maxx;
         }
      }
   }
   free(sumX);
   free(sumY);
   free(yy);
   return mask;
} /* end of midiCreateMask */


/* *********************************************************
   *  use top and bottom rows to estimate sky in each
   *  window and subtract linear fit
   ********************************************************** */

void deSky (minrtsFrameData *data, int y1, int y2) {
   int ix, iy;
   float dy, x1, x2;
   float topSky, bottomSky;
   float *yLine;
   int nx;
   int iregion;

   for (iregion=0; iregion<data->nregion; iregion++) {
      nx = data->nx[iregion];

/*
      average a few rows and do a linear fit
*/
      dy = y2 - y1;
      for (ix=0; ix<nx; ix++) {
         for (iy =-1; iy<=1; iy++) {
            topSky    = topSky +    data->data[iregion]
               [ix + (iy+y2)*(long)nx];
            bottomSky = bottomSky + data->data[iregion]
               [ix + (iy+y1)*(long)nx];
         }
         topSky    = topSky/3.;
         bottomSky = bottomSky/3.;
/*
      linear fit and subtraction
*/
         for (iy=0; iy<data->ny[iregion]; iy++) {
            x1 = ((float)iy - y1)/dy;
            x2 = ((float)iy - y2)/dy;
            data->data[iregion][ix+iy*nx] = data->data[iregion][ix+iy*nx]-
               topSky*x1 + bottomSky*x2;
         } /* loop over y */
      }    /* loop over x */
   } /* loop over regions */
}

int main (int argc, char** argv) {

   char opt1[12];
   char gris[12];
   char shutter[12];
   float dit;

   char **fileList;
   int nFiles;
   int iShutter;
   int i,ierr;
   int nx, ny;
   long npix;
   long iPix;
   int  ix, iy;
   float xx, mm;
   long  ii;
   int iKey;
   int nCycle;

   int chan1, chan2;

   char outFileName[130];

   fitsfile *inFile;

   imaging_fdata_table_def *odata;
   imaging_detector_table_def *DetectorDef;
   imaging_detector_table_def *CoordDef;
   imaging_fdata_row *outData;
   int nRegion;


   minrtsFrameData *Target;
   minrtsFrameData *Sky;
   minrtsFrameData *Mask;
   minrtsFrameData *Phot;
   minrtsFrameData *Cross ;


   float sigmaY ; 
   float *ATotal, *BTotal, *ABTotal;
   float *AMask, *BMask, *ABMask;

   float t1, t2;
   float p02, p13;

   char *bang="!";
   char *ABshutter = "ABOPEN";
   char *sciPhot  = "SCI_PHOT";
   fitsfile *outFile = NULL;
   int nDrop = 0;
   int one = 1;

   ierr = 0;
/*
      see if sigmaY is specified externally
*/
   sigmaY = SIGMAY;
   if (argc > 6) sscanf(argv[6], "%f", &sigmaY);
   printf("nDrop = %i\n", nDrop);
   printf("sigmaY = %f\n", sigmaY);

/*        
       break up file list and print what you find 
       then read and unchop files
*/
   fileList = oirStrSplit(argv[1], &nFiles);

   i = fits_open_file(&inFile, fileList[0], READONLY, &ierr) ;
   if (ierr != 0) {
      printf("cfitsio error %i opening file %s\n", ierr, fileList[0]);
      printf("%s \n",strerror(errno));
      return 1;
   }
   ierr = 0;
   i = fits_read_key(inFile, TSTRING, "HIERARCH ESO INS SHUT NAME", 
      shutter, 0, &ierr);
   if (strncmp(shutter, ABshutter, 6)!=0) {
      fprintf(stderr, "I expect shutter ABOPEN not %s \n", shutter);
      return 1;
   }
   i = fits_read_key(inFile, TFLOAT, "HIERARCH ESO DET DIT", &dit, 
      0, &ierr);
   printf("dit = %f\n", dit);
   ierr = 0;
   i = fits_read_key(inFile, TSTRING, "HIERARCH ESO INS OPT1 NAME", 
      opt1,0, &ierr);
   if (strcmp(opt1, sciPhot) != 0) {
      fprintf(stderr, "I expect opt1 to be SCI_PHOT not %s \n", opt1);
      return 1;
   }
   i = fits_read_key(inFile, TSTRING, "HIERARCH ESO INS GRIS NAME", 
      gris,0,&ierr);
   printf("disperser is = %s\n", gris);
/*
      move to detector section
*/
   DetectorDef = read_imaging_detector_table(inFile, &ierr);
   if (ierr !=0) {
      printf("error %i opening detector table in %s\n", 
         ierr, fileList[0]);
      return 1;
   }
   nRegion = DetectorDef->nregion;

   i = fits_close_file(inFile, &ierr);

/*
       get detector table with coordinate distortion info
*/

   i = fits_open_file(&inFile, argv[2], READONLY, &ierr) ;
   if (ierr != 0) {
      printf("cfitsio error %i opening coord file %s\n", ierr, argv[2]);
      printf("%s \n",strerror(errno));
      return 1;
   }
   CoordDef = read_imaging_detector_table(inFile, &ierr);
   if (ierr !=0) {
      printf("error %i opening Coord detector table in %s\n", 
         ierr, argv[2]);
      return 1;
   }
   printf("using optical distortion data from %s\n", argv[2]);
   i = fits_close_file(inFile, &ierr);

/*
          create FrameData structures to hold data
*/
   Target = minrtsFrameDataMake(DetectorDef);
   Sky    = minrtsFrameDataMake(DetectorDef);
   Mask   = minrtsFrameDataMake(DetectorDef);
   Phot   = minrtsFrameDataMake(DetectorDef);
/*
          now read in the real data
*/
   printf("Found %i data Files\n", nFiles);
   for (i=0; i<nFiles; i++) printf("%s \n", fileList[i]);
   
   printf("Averaging Target and Sky Frames\n");
   minrtsFrameDataUnChop (fileList, nFiles, nDrop,
      Target, Sky, &ierr);
   if (ierr != 0) {
      fprintf(stderr,"Failed to process Data: error: %i\n", ierr);
      fprintf(stderr,"%s \n", strerror(errno));
      return 1;
   }
   printf("Found %7.0f Target and %7.0f Sky frames\n", Target->weight, Sky->weight);
   oirKillFileList(fileList, nFiles);

#ifdef DEBUG 
minrtsFrameDataWriteFits(Target, "target.fits", &ierr);
minrtsFrameDataWriteFits(Sky, "sky.fits", &ierr);
#endif
/*
       subtract sky and divide by dit
*/
   minrtsFrameDataSub(Target, Target, Sky);

#ifdef DEBUG 
minrtsFrameDataWriteFits(Target, "diff.fits", &ierr);
#endif

   if (dit > 0.) minrtsFrameDataScalarMul(Target, 1./dit);

/*
      deSky and Rectify images
*/
   deSky(Target, yBottom, yTop);
#ifdef DEBUG 
minrtsFrameDataWriteFits(Target, "desky.fits", &ierr);
#endif
   midiRectXYWindow(Target, CoordDef, 2, &ierr);
#ifdef DEBUG 
minrtsFrameDataWriteFits(Target, "rect.fits", &ierr);
#endif
   nx = Target->nx[0];
   ny = Target->ny[0];
/*
      allocate output arrays
*/
   ATotal = (float *)malloc(nx* sizeof(float));
   BTotal = (float *)malloc(nx* sizeof(float));
   ABTotal= (float *)malloc(nx* sizeof(float));
   AMask  = (float *)malloc(nx* sizeof(float));
   BMask  = (float *)malloc(nx* sizeof(float));
   ABMask = (float *)malloc(nx* sizeof(float));
/*
      duplicate A-B photometry before applying cross coupling coeff
*/
   for (ix=0;ix<nx;ix++) for (iy=0;iy<ny;iy++) {
      i = ix + iy*nx;
      Phot->data[0][i] = Target->data[3][i];
      Phot->data[1][i] = Target->data[3][i];
      Phot->data[2][i] = Target->data[0][i];
      Phot->data[3][i] = Target->data[0][i];
   }
/*
      create mask arrays and save
*/
   Mask = midiCreateMask(Target, sigmaY);
   minrtsFrameDataWriteFits(Mask, argv[5], &ierr);
   if (ierr != 0) {
      fprintf(stderr, "error %i writing mask data to file %s\n", ierr, argv[5]);
      fprintf(stderr, "%s \n", strerror(errno));
      return 1;
   }
   printf("writing mask file into %s\n", argv[5]);
/*
      get cross coupling coefficients
      multiply Phot by Cross
*/
   Cross = minrtsFrameDataReadFitsRow(argv[3],1,&ierr);
   if (ierr != 0) {
      fprintf(stderr, "error %i getting kappa coeff from file %s\n", ierr, argv[3]);
      fprintf(stderr, "%s \n", strerror(errno));
      return 1;
   }
   printf("getting kappa coefficients from  %s\n", argv[3]);
   for (i=0;i<4;i++) for(ix=0;ix<nx;ix++) for (iy=0;iy<ny;iy++) 
         Phot->data[i][ix+iy*nx] = Phot->data[i][ix+iy*nx]*Cross->data[i][ix];
#ifdef DEBUG 
minrtsFrameDataWriteFits(Phot, "Phot.fits", &ierr);
#endif
#ifdef RECAL
/*
      modify effective kappa coeff to make sum of Phot[0+2] = Target[1]
      modify effective kappa coeff to make sum of Phot[1+3] = Target[3]
*/
   for (ix=0;ix<nx;ix++) {
      p02 = 0.;
      t1  = 0.;
      p13 = 0.;
      t2  = 0.;
      for (iy=yBottom;iy<yTop;iy++) {
         i = ix + iy * nx;
         p02 = p02+Phot->data[0][i] + Phot->data[2][i];
         p13 = p13+Phot->data[1][i] + Phot->data[3][i];
         t1  = t1 +Target->data[1][i];
         t2  = t2 +Target->data[2][i];
      }
      if (p02 > 0.) t1 = t1/p02; else t1 = 1.;
      if (p13 > 0.) t2 = t2/p13;else t2 = 1.;
      for (iy=0;iy<ny;iy++) {
         i = ix + iy * nx;
         Phot->data[0][i] = Phot->data[0][i] * t1;
         Phot->data[2][i] = Phot->data[2][i] * t1;
         Phot->data[1][i] = Phot->data[1][i] * t2;
         Phot->data[3][i] = Phot->data[3][i] * t2;
      }
   }
#ifdef DEBUG 
minrtsFrameDataWriteFits(Phot, "PhotC.fits", &ierr);
#endif
#endif
/*
      compute  totals
*/
   for (ix=0;ix<nx;ix++) {
      ATotal[ix]  = 0.;
      BTotal[ix]  = 0.;
      ABTotal[ix] = 0.;
      AMask[ix]  = 0.;
      BMask[ix]  = 0.;
      ABMask[ix] = 0.;
      for (iy=yBottom;iy<yTop;iy++) {
         i = ix + iy*nx;
         xx = Phot->data[0][i] + Phot->data[1][i];
         ATotal[ix] = ATotal[ix] + xx;
         AMask[ix]  = AMask[ix] + xx * Mask->data[1][i];
         xx = Phot->data[2][i] + Phot->data[3][i];
         BTotal[ix] = BTotal[ix] + xx;
         BMask[ix]  = BMask[ix] + xx * Mask->data[1][i];
         xx = Phot->data[0][i]*Phot->data[2][i];
         if (xx > 0.) {
            xx = sqrt(xx);
            ABTotal[ix] = ABTotal[ix] + xx;
            ABMask[ix] = ABMask[ix] + xx * Mask->data[1][i];
         }
         xx = Phot->data[1][i]*Phot->data[3][i];
         if (xx > 0.) {
            xx = sqrt(xx);
            ABTotal[ix] = ABTotal[ix] + xx;
            ABMask[ix] = ABMask[ix] + xx * Mask->data[1][i];
         }
      }  /* sum over y */
   } /* loop over x */
/*
      prepare output file
*/
   strcpy(outFileName, bang);
   strcat(outFileName, argv[4]);

   fits_create_file (&outFile, outFileName, &ierr);
   if (ierr !=0) {
      fprintf (stderr, "error opening fits output %s %i\n",argv[4], ierr);
      printf("%s \n",strerror(errno));
      return 1;
   }
   printf("writing output photometry into %s\n",argv[4]);
   DetectorDef->nregion = 1;
   DetectorDef->naxis[0][1] = 1;
   iKey = table_find_key(DetectorDef->def, "NREGION");
   table_put_keyword(DetectorDef->def, iKey, (void *)&one, &ierr);

   odata = make_imaging_fdata_table_def (outFile, DetectorDef, 1, &ierr);
   if (ierr !=0) {
      fprintf (stderr, "error creating data definition\n");
      return 1;
   }
   outData = make_imaging_fdata_row(odata);

   for (ix=0;ix<nx;ix++) outData->data[0][ix] = ATotal[ix];
   i = write_imaging_fdata_table(outFile, odata, outData, 1, &ierr);

   for (ix=0;ix<nx;ix++) outData->data[0][ix] = BTotal[ix];
   i = write_imaging_fdata_table(outFile, odata, outData, 2, &ierr);

   for (ix=0;ix<nx;ix++) outData->data[0][ix] = ABTotal[ix];
   i = write_imaging_fdata_table(outFile, odata, outData, 3, &ierr);

   for (ix=0;ix<nx;ix++) outData->data[0][ix] = AMask[ix];
   i = write_imaging_fdata_table(outFile, odata, outData, 4, &ierr);

   for (ix=0;ix<nx;ix++) outData->data[0][ix] = BMask[ix];
   i = write_imaging_fdata_table(outFile, odata, outData, 5, &ierr);

   for (ix=0;ix<nx;ix++) outData->data[0][ix] = ABMask[ix];
   i = write_imaging_fdata_table(outFile, odata, outData, 6, &ierr);

   for (ix=0;ix<nx;ix++) outData->data[0][ix] = 0.;/* ATotal[ix];*/
   i = write_imaging_fdata_table(outFile, odata, outData, 7, &ierr);

   for (ix=0;ix<nx;ix++) outData->data[0][ix] = 0.;/*BTotal[ix];*/
   i = write_imaging_fdata_table(outFile, odata, outData, 8, &ierr);

   for (ix=0;ix<nx;ix++) outData->data[0][ix] = 0.;/*ABTotal[ix];*/
   i = write_imaging_fdata_table(outFile, odata, outData, 9, &ierr);

   for (ix=0;ix<nx;ix++) outData->data[0][ix] = 0.;/*AMask[ix];*/
   i = write_imaging_fdata_table(outFile, odata, outData, 10, &ierr);

   for (ix=0;ix<nx;ix++) outData->data[0][ix] = 0.;/*BMask[ix];*/
   i = write_imaging_fdata_table(outFile, odata, outData, 11, &ierr);

   for (ix=0;ix<nx;ix++) outData->data[0][ix] = 0.;/*ABMask[ix];*/
   i = write_imaging_fdata_table(outFile, odata, outData, 12, &ierr);

   i = close_imaging_fdata_table(outFile, odata, &ierr);
   kill_imaging_fdata_row(outData);

   minrtsFrameDataKill(Target);
   minrtsFrameDataKill(Sky);
   minrtsFrameDataKill(Mask);
   minrtsFrameDataKill(Phot);
   minrtsFrameDataKill(Cross);
   free(AMask);
   free(BMask);
   free(ABMask);
   free(ATotal);
   free(BTotal);
   free(ABTotal);
   return 0;
}
